local lib = LibStub:NewLibrary("LibGuildStorage", 1)

local ADDON_MESSAGE_PREFIX = "Guild Storage"
RegisterAddonMessagePrefix(ADDON_MESSAGE_PREFIX)

if not lib then return end

local GUILDFRAMEVISIBLE = false
local OUTSIDERSENABLED = false

local CallbackHandler = LibStub("CallbackHandler-1.0")
if not lib.callbacks then
	lib.callbacks = CallbackHandler:New(lib)
end
local callbacks = lib.callbacks

local AceHook = LibStub("AceHook-3.0")
AceHook:Embed(lib)
lib:UnhookAll()

if lib.frame then
	lib.frame:UnregisterAllEvents()
	lib.frame:SetScript("OnEvent", nil)
	lib.frame:SetScript("OnUpdate", nil)
else
	lib.frame = CreateFrame("Frame", "1" .. "_Frame")
end
local frame = lib.frame
frame:Show()
frame:SetScript("OnEvent",
function(self, event, ...)
	lib[event](lib, ...)
end)

local SendAddonMessage = _G.SendAddonMessage
if ChatThrottleLib then
	SendAddonMessage = function(...)
		ChatThrottleLib:SendAddonMessage(
		"ALERT", ADDON_MESSAGE_PREFIX, ...)
	end
end

local SetState

local state = "STALE_WAITING_FOR_ROSTER_UPDATE"
local initialized
local index
local cache = {}
local pending_note = {}
local guild_info = ""

function lib:GetNote(name)
	local e = cache[name]
	if e then return e.note end
end

function lib.GetIndex(name)
	
end

function lib:GetClass(name)
	local e = cache[name]
	if e then return e.class end
end

function lib:GetRank(name)
	local e = cache[name]
	if e then return e.rank end
end

function lib:GetClassName(name)
	local e = cache[name]
	if e then return e.className end
end		
				
function lib:GetLevel(name)
	local e = cache[name]
	if e then return e.level end
end		
			
function lib:GetZone(name)
	local e = cache[name]
	if e then return e.zone end
end		
			
function lib:GetAchievementPoints(name)
	local e = cache[name]
	if e then return e.achievementPoints end
end		
			
function lib:GetOnline(name)
	local e = cache[name]
	if e then return e.online end
end			
		
function lib:GetPubNote(name)
	local e = cache[name]
	if e then return e.pubNote end
end						

function lib:GetGuildInfo()
	return guild_info
end

function lib:IsCurrentState()
	return state == "CURRENT"
end

-- This function allows users to enable or disable the outsiders patch
function lib:SetOutsidersEnabled(isOutsidersEnabled)
	-- Dont do anything if the boolean is the same
	if (OUTSIDERSENABLED == isOutsidersEnabled) then
		return
	end

	OUTSIDERSENABLED = isOutsidersEnabled

	--print("outsider changed, now is ", OUTSIDERSENABLED)
	-- Force reloading of guildnotes
	index = nil
	SetState("STALE")
end

--
-- Event handlers
--
frame:RegisterEvent("PLAYER_GUILD_UPDATE")
frame:RegisterEvent("GUILD_ROSTER_UPDATE")
frame:RegisterEvent("CHAT_MSG_ADDON")
frame:RegisterEvent("PLAYER_ENTERING_WORLD")

function lib:CHAT_MSG_ADDON(prefix, msg, type, sender)
	--print("CHAT_MSG_ADDON: %s, %s, %s, %s", prefix, msg, type, sender)
	if prefix ~= MAJOR_VERSION or sender == UnitName("player") then return end
	if msg == "CHANGES_PENDING" then
		SetState("REMOTE_FLUSHING")
	elseif msg == "CHANGES_FLUSHED" then
		SetState("STALE_WAITING_FOR_ROSTER_UPDATE")
	end
end

function lib:PLAYER_GUILD_UPDATE()
	if IsInGuild() then
		frame:Show()
	else
		frame:Hide()
	end
	SetState("STALE_WAITING_FOR_ROSTER_UPDATE")
end

function lib:PLAYER_ENTERING_WORLD()
	lib:PLAYER_GUILD_UPDATE()
end

function lib:GUILD_ROSTER_UPDATE(loc)
	--print("GUILD_ROSTER_UPDATE(%s)", tostring(loc))
	if loc then
		SetState("FLUSHING") -- SetState("STALE_WAITING_FOR_ROSTER_UPDATE")
	else
		if state ~= "UNINITIALIZED" then
			SetState("STALE")
			index = nil
		end
	end
end

--
-- Locally defined functions
--

local valid_transitions = {
	UNINITIALIZED = {
		CURRENT = true,
	},
	STALE = {
		CURRENT = true,
		REMOTE_FLUSHING = true,
		STALE_WAITING_FOR_ROSTER_UPDATE = true,
	},
	STALE_WAITING_FOR_ROSTER_UPDATE = {
		STALE = true,
		FLUSHING = true,
	},
	CURRENT = {
		FLUSHING = true,
		REMOTE_FLUSHING = true,
		STALE = true,
	},
	FLUSHING = {
		STALE_WAITING_FOR_ROSTER_UPDATE = true,
	},
	REMOTE_FLUSHING = {
		STALE_WAITING_FOR_ROSTER_UPDATE = true,
	},
}

function SetState(new_state)
	if state == new_state then return end

	if not valid_transitions[state][new_state] then
		--print("Ignoring state change %s -> %s", state, new_state)
		return
		else
		--print("StateChanged: %s -> %s", state, new_state)
		state = new_state
		if new_state == FLUSHING then
			SendAddonMessage("CHANGES_PENDING", "GUILD")
		end
		callbacks:Fire("StateChanged")
	end
end

local function ForceShowOffline()
	-- We need to always show offline members in the roster otherwise this
	-- lib won't work.

	if GUILDFRAMEVISIBLE then
		return true
	end

	SetGuildRosterShowOffline(true)

	return false
end

local function Frame_OnUpdate(self, elapsed)
	local startTime = debugprofilestop()
	if ForceShowOffline() then
		return
	end

	if state == "CURRENT" then
		return
	end

	if state == "STALE_WAITING_FOR_ROSTER_UPDATE" then
		GuildRoster()
		return
	end

	local num_guild_members = GetNumGuildMembers()

	-- Sometimes GetNumGuildMembers returns 0. In this case return now,
	-- so that we call it again and get a proper value.
	if num_guild_members == 0 then return end

	if not index or index >= num_guild_members then
		index = 1
	end

	-- Check guild info for changes.
	if index == 1 then
		local new_guild_info = GetGuildInfoText() or ""
		if new_guild_info ~= guild_info then
			guild_info = new_guild_info
			callbacks:Fire("GuildInfoChanged", guild_info)
		end
	end

	-- Read up to 100 members at a time.
	local last_index = math.min(index + 100, num_guild_members)
	if not initialized then last_index = num_guild_members end
	--print("Processing from %d to %d members", index, last_index)

	for i = index, last_index do

		local name, rank, _, level, className, zone, pubNote, note, online, _, class, achievementPoints = GetGuildRosterInfo(i)
		-- We use full names including the '-server' portion
		local name = Ambiguate(name, "mail")

		-- Start of outsiders patch
		if OUTSIDERSENABLED then
			local extName = strmatch(pubNote, 'ext:%s-(%S+)%s-')
			local holder
			if extName then
				-- the name is now the note and the external name is the new name.
				local entry = cache[extName]
				if not entry then
					entry = {}
					cache[extName] = entry
				end

				entry.className = className
				entry.level = level
				entry.zone = zone
				entry.achievementPoints = achievementPoints
				entry.online = online
				entry.pubNote = pubNote

				--print("Entry " .. holder .. " is " .. extName)
				-- Mark this note as seen
				entry.seen = true
				if entry.note ~= holder then
					entry.note = holder
					local _, unitClass = UnitClass(extName)
					entry.rank = "Outsider("..name..")"
					entry.class = unitClass or class
					if initialized then
						callbacks:Fire("GuildNoteChanged", extName, holder)
					end
					if entry.pending_note then
						callbacks:Fire("InconsistentNote", extName, holder, entry.note, entry.pending_note)
					end
				end

				if entry.pending_note then
					GuildRosterSetOfficerNote(i, entry.pending_note)
					entry.pending_note = nil
				end
			end
		end -- if OUTSIDERSENABLED

		if name then
			local entry = cache[name]
			local pending = pending_note[name]
			if not entry then
				entry = {}
				cache[name] = entry
			end

			entry.rank = rank
			entry.class = class
			
			entry.className = className
			entry.level = level
			entry.zone = zone
			entry.achievementPoints = achievementPoints
			entry.online = online
			entry.pubNote = pubNote

			-- Mark this note as seen
			entry.seen = true
			if entry.note ~= note then
				entry.note = note
				-- We want to delay all GuildNoteChanged calls until we have a
				-- complete view of the guild, otherwise alts might not be
				-- rejected (we read alts note before we even know about the
				-- main).
				if initialized then
					callbacks:Fire("GuildNoteChanged", name, note)
					callbacks:Fire("GuildMemberChanged", level, name, className, class, zone, pubNote, note, achievementPoints, online )
				end
				if pending then
					callbacks:Fire("InconsistentNote", name, note, entry.note, pending)
				end
			end

			if pending then
				GuildRosterSetOfficerNote(i, pending)
				pending_note[name] = nil
			end
		end
	end
	index = last_index
	if index >= num_guild_members then
		-- We are done, we need to clear the seen marks and delete the
		-- unmarked entries. We also fire events for removed members now.
		for name, t in pairs(cache) do
			if t.seen then
				t.seen = nil
			else
				cache[name] = nil
				callbacks:Fire("GuildNoteDeleted", name)
			end
		end

		if not initialized then
			-- Now make all GuildNoteChanged calls because we have a full
			-- state.
			for name, t in pairs(cache) do
				callbacks:Fire("GuildNoteChanged", name, t.note)				
				callbacks:Fire("GuildMemberChanged", t.level, name, t.className, t.class, t.zone, t.pubNote, t.note, t.achievementPoints, t.online )
			end
			initialized = true
			callbacks:Fire("StateChanged")
		end
		if state == "STALE" then
			SetState("CURRENT")
		elseif state == "FLUSHING" then
			if not next(pending_note) then
				SetState("STALE_WAITING_FOR_ROSTER_UPDATE")
				SendAddonMessage("CHANGES_FLUSHED", "GUILD")
			end
		end
	end
--print(tostring(debugprofilestop() - startTime).."ms for LibGuildStorage:OnUpdate")
end

-- Disable updates when the guild roster is open.
-- This is a temporary hack until we get a better location for data storage
lib:RawHook("GuildFrame_LoadUI", function(...)
	lib.hooks.GuildFrame_LoadUI(...)
	lib:RawHookScript(GuildRosterFrame, "OnShow", function(frame, ...)
		GUILDFRAMEVISIBLE = true
		if GuildRosterShowOfflineButton then
			GuildRosterShowOfflineButton:Enable()
		end
		lib.hooks[frame].OnShow(frame, ...)
	end)
	lib:RawHookScript(GuildRosterFrame, "OnHide", function(frame, ...)
		GUILDFRAMEVISIBLE = false
		lib.hooks[frame].OnHide(frame, ...)
		SetGuildRosterShowOffline(true)
	end)
	lib:Unhook("GuildFrame_LoadUI")

	SetGuildRosterShowOffline(true)
end, true)


ForceShowOffline()
frame:SetScript("OnUpdate", Frame_OnUpdate)
GuildRoster()
